-- Generic routines for using Dynamic Link Libraries (3rd release)
-- Should work on all (x86) platforms
--
-- Use the same way you would use the standard dll.e
--
-- /Mic, 2002
--

without warning

include file.e
include machine.e
include asm.e


global constant 
	 C_CHAR    = #01000001,
	 C_UCHAR   = #02000001,
	 C_SHORT   = #01000002,
	 C_USHORT  = #02000002,
	 C_INT     = #01000004,
	 C_UINT    = #02000004,
	 C_LONG    = C_INT,
	 C_ULONG   = C_UINT,
	 C_POINTER = C_ULONG,
	 C_FLOAT   = #03000004,
	 C_DOUBLE  = #03000008

global constant GENDLL_STDCALL = 0,
		GENDLL_CDECL = 1,
		GENDLL_FASTCALL = 2
		
constant DLL_PROCESS_DETACH = 0,
	 DLL_PROCESS_ATTACH = 1,
	 DLL_THREAD_ATTACH = 2,
	 DLL_THREAD_DETACH = 3


constant IDT_LUT_RVA = 1,
	 IDT_TIME = 2,
	 IDT_FWD = 3,
	 IDT_NAME_RVA = 4,
	 IDT_ADT_RVA = 5,
	 IDT_LUT = 6

constant SECT_NAME = 1,
	 SECT_VIRTUAL_SIZE = 2,
	 SECT_RVA = 3,
	 SECT_PHYS_SIZE = 4,
	 SECT_PHYS_OFFS = 5,
	 SECT_FLAGS = 6

constant IMG_SCN_CNT_CODE = #00000020,			-- The section contains code
	 IMG_SCN_CNT_INIT_DATA = #00000040,		-- The section contains data
	 IMG_SCN_CNT_UNINIT_DATA = #00000080,		-- The section contains uninitialized data
	 IMG_SCN_MEM_EXECUTE = #20000000,		-- The memory is executable
	 IMG_SCN_MEM_READ = #40000000			-- The memory is readable
	 
constant DLL_TEMPLATE = {0,0,0,0,0}

constant DLL_BASE_ADDRESS = 1,
	 DLL_EAT = 2,
	 DLL_ENT = 3,
	 DLL_EOT = 4,
	 DLL_DEFS = 5

constant gendll_regs = {"eax","ebx","ecx","edx","esi","edi"}


sequence loaded_dlls,
	 gendll_fundefs,
	 gendll_ordinals,
	 gendll_name_filters

integer gendll_call_mode
	 
loaded_dlls = {}
gendll_fundefs = {}
gendll_ordinals = {}
gendll_call_mode = GENDLL_STDCALL



if platform()=1 then
	gendll_name_filters = {{"kernel32.dll"},{"dosknl32.dll"}}

elsif platform()=3 then
	gendll_name_filters = {{"kernel32.dll"},{"./linknl32.dll"}}

else
	-- If someone tries to load dll in [1], load the dll in [2] instead
	gendll_name_filters = {{"kernel32.dll"},{"genknl32.dll"}}	
end if



constant LibMain_data = allocate(8)
constant call_LibMain = get_asm(
"pusha "&
"mov eax,[reason] "&
"push eax "&
"push eax "&
"push eax "&
"mov ebp,[LibMain_addr] "&
"call near ebp "&
"mov [result],eax "&
"popa "&
"ret")
resolve_param("LibMain_addr",LibMain_data)
resolve_param("reason",LibMain_data+4)
resolve_param("result",LibMain_data+4)


function gendll_word(sequence s)
	return s[1]+(s[2]*256)
end function

function gendll_dword(sequence s)
	return s[1]+(s[2]*#100)+(s[3]*#10000)+(s[4]*#1000000)
end function

procedure gendll_poke2(atom addr,integer wrd)
	poke(addr,{and_bits(wrd,#FF),floor(wrd/256)})
end procedure


function get_bytes(integer fn, integer nbytes)
	sequence s
	integer i

	s = {}
	for j=1 to nbytes do
		i = getc(fn)
		if i=-1 then
			exit
		end if
		s &= i
	end for
	return s
end function

	
-- Convert a Relative Virtual Address to a filepointer
function rva_to_filepointer(integer rva,sequence sectInfo)

	integer temp

	temp = 1
	while (rva < sectInfo[temp][SECT_RVA]) or
	      (rva >= sectInfo[temp][SECT_RVA] + sectInfo[temp][SECT_VIRTUAL_SIZE]) do
		temp += 1
	end while
	
	if rva>sectInfo[temp][SECT_RVA] then
		return sectInfo[temp][SECT_PHYS_OFFS] + rva - sectInfo[temp][SECT_RVA] 
	end if
	
	return sectInfo[temp][SECT_PHYS_OFFS]

end function


-- Is the library already loaded ?
function gendll_lib_is_loaded(sequence libname)
	sequence s,lowlibn
	
	lowlibn = lower(libname)
	
	for i=1 to length(loaded_dlls) do
		s = loaded_dlls[i]
		if length(s) then
			if equal(lower(s[DLL_ENT][1]), lowlibn) then
				return i
			end if
		end if
	end for
	
	return 0
end function


function gendll_func_is_def_byname(sequence func_name)
	sequence s
	
	for i=1 to length(loaded_dlls) do
		s = loaded_dlls[i]
		if length(s) then
			for j=1 to length(s[DLL_ENT])-1 do
				if equal(s[DLL_ENT][j+1], func_name) then
					return i
				end if
			end for
		end if
	end for
	
	return 0
end function


function gendll_get_proc_address(integer lib, sequence func_name)
	sequence s
	
	s = loaded_dlls[lib]
	if length(s) then
		for j=1 to length(s[DLL_ENT])-1 do
			if equal(s[DLL_ENT][j+1], func_name) then
				return s[DLL_BASE_ADDRESS]+s[DLL_EAT][j]
			end if
		end for
	end if
	
	return 0
end function




global procedure add_dll_name_filter(sequence realname,sequence usedname)
	gendll_name_filters[1] = append(gendll_name_filters[1],realname)
	gendll_name_filters[2] = append(gendll_name_filters[2],usedname)
end procedure




global function open_dll(sequence fname)

	sequence DOSh,
	  	 WINh,
		 SECTh,
		 SECT,
		 EAT,		-- Export Address Table
		 EDT,		-- Export Directory Table
		 ENP,		-- Export Name table Pointers
		 EOT,		-- Export Ordinal Table
		 ENT,		-- Export Name Table
		 STDh,
		 IDT,		-- Import Directory Table
		 NTh,
		 IDD,
		 stemp

	integer fn,temp,
		numSects,
		sectAlign,
		fileAlign,
		optHdrSize,
		headerSize,
		stackRes,stackComm,
		heapRes,heapComm,
		dllChar,
		numRVAandSizes,
		impTblAddr,impTblSize,
		expTblAddr,expTblSize,
		relTblAddr,relTblSize,
		lMinor,lMajor,PEtype,
	     	relType,relOffs

	atom entryRVA,
	     imageBase,
	     codeBase,
	     dataBase,
	     codeSize,
	     imageSize,
	     initDataSize,uninitDataSize,
	     imagePtr,
	     pageRVA,
	     relBlkSize,
	     relDelta,
	     atemp,
	     implib


	temp = gendll_lib_is_loaded(fname)
	if temp>0 then
		return temp
	end if
	
	temp = find(lower(fname),gendll_name_filters[1])
	if temp>0 then
		fname = gendll_name_filters[2][temp]
	end if
	
	--puts(1,"Loading "&fname&"\n")
	
	fn = open(fname,"rb")
	if fn=-1 then
		--puts(1,"Can\'t find "&fname)
		--abort(0)
		return -1
	end if

	DOSh = get_bytes(fn, #1C)
	if length(DOSh)<#1C then
		--puts(1,"Error while reading file!")
		close(fn)
		--abort(0)
		return -1
	end if


	if gendll_word(DOSh[#19..#1A])<#40 then
		--puts(1,"No Windows header found!")
		close(fn)
		--abort(0)
		return -1
	end if

	-- Move to Windows PE header
	if seek(fn,#3C) then end if
	temp = getc(fn)
	if seek(fn,temp) then end if


	WINh = get_bytes(fn,24)
	if length(WINh)<24 then
		--puts(1,"Error while reading file!")
		close(fn)
		--abort(0)
		return -1
	end if
	
	if not equal(WINh[1..2],"PE") then
		--puts(1,"Bad signature!\n")
		close(fn)
		--abort(0)
		return -1
	end if

	WINh = WINh[5..length(WINh)]


	numSects = gendll_word(WINh[3..4])
	if not and_bits(gendll_word(WINh[19..20]),#2000) then
		--puts(1,"Image is not a DLL!\n")
		close(fn)
		--abort(0)
		return -1
	end if


	optHdrSize = gendll_word(WINh[17..18])
	if optHdrSize=0 then
		--puts(1,"Not enough information to complete load!")
		close(fn)
		--abort(0)
		close(fn)
	end if
	
	WINh = get_bytes(fn,optHdrSize)
	if length(WINh)<optHdrSize then
		--puts(1,"Error while reading file!")
		close(fn)
		--abort(0)
		return -1
	end if
	
	PEtype = gendll_word(WINh[1..2])
	if PEtype != #10B then
		--puts(1,"Unsupported PE format!")
		close(fn)
		--abort(0)
		return -1
	end if

--#################### Read the standard COFF header ####################	
	
	STDh = WINh[1..28]
	codeSize = gendll_dword(STDh[5..8])
	initDataSize = gendll_dword(STDh[9..12])
	uninitDataSize = gendll_dword(STDh[13..16])
	entryRVA = gendll_dword(STDh[17..20])
	codeBase = gendll_dword(STDh[21..24])
	dataBase = gendll_dword(STDh[25..28])

--#################### Read the NT header ####################	

	NTh = WINh[1..96]
	imageBase = gendll_dword(NTh[29..32])
	sectAlign = gendll_dword(NTh[33..36])
	fileAlign = gendll_dword(NTh[37..40])
	imageSize = gendll_dword(NTh[57..60])
	headerSize = gendll_dword(NTh[61..64])
	
	dllChar = gendll_word(NTh[71..72])
	stackRes = gendll_dword(NTh[73..76])
	stackComm = gendll_dword(NTh[77..80])
	heapRes = gendll_dword(NTh[81..84])
	heapComm = gendll_dword(NTh[85..88])
	numRVAandSizes = gendll_dword(NTh[93..96])
	

	expTblAddr = gendll_dword(WINh[97..100])
	expTblSize = gendll_dword(WINh[101..104])
	impTblAddr = gendll_dword(WINh[105..108])
	impTblSize = gendll_dword(WINh[109..112])
	relTblAddr = gendll_dword(WINh[137..140])
	relTblSize = gendll_dword(WINh[141..144])
	
	--printf(1,"exp tbl addr: %04xh\nsize: %d\n",{expTblAddr,expTblSize})
	--printf(1,"rel tbl addr: %04xh\nsize: %d\n",{relTblAddr,relTblSize})
	
	SECT = repeat(repeat(0,6),numSects)

	for i=1 to numSects do

		SECTh = get_bytes(fn,40)
		SECT[i][SECT_NAME] = {}
	
		-- Get the name of the section (remove trailing null bytes)
		for j=1 to 8 do
			if SECTh[j] != 0 then
				SECT[i][SECT_NAME] &= SECTh[j]
			end if
		end for

		SECT[i][SECT_VIRTUAL_SIZE] = gendll_dword(SECTh[9..12])
		SECT[i][SECT_RVA] = gendll_dword(SECTh[13..16])
		SECT[i][SECT_PHYS_SIZE] = gendll_dword(SECTh[17..20])
		SECT[i][SECT_PHYS_OFFS] = gendll_dword(SECTh[21..24])
		SECT[i][SECT_FLAGS] = gendll_dword(SECTh[37..40])

	end for


	-- Allocate memory for the image
	imagePtr = allocate(imageSize)
	if imagePtr<=0 then
		return -1
	end if


	
--#################### Copy code and data sections ####################	

	for i=1 to numSects do
		if length(SECT[i]) then
			if seek(fn,SECT[i][SECT_PHYS_OFFS]) then end if
			poke(imagePtr+SECT[i][SECT_RVA],get_bytes(fn,SECT[i][SECT_VIRTUAL_SIZE]))
		end if
	end for

	
--#################### Process export table ####################	

	if expTblSize>0 then

		if seek(fn,rva_to_filepointer(expTblAddr, SECT)) then end if
		
		stemp = get_bytes(fn,40)
		EDT = repeat(0,11)
		EDT[1] = gendll_dword(stemp[1..4])
		EDT[2] = gendll_dword(stemp[5..8])
		EDT[3] = gendll_word(stemp[9..10])
		EDT[4] = gendll_word(stemp[11..12])
		EDT[5] = gendll_dword(stemp[13..16])
		EDT[6] = gendll_dword(stemp[17..20])
		EDT[7] = gendll_dword(stemp[21..24])
		EDT[8] = gendll_dword(stemp[25..28])
		EDT[9] = gendll_dword(stemp[29..32])
		EDT[10] = gendll_dword(stemp[33..36])
		EDT[11] = gendll_dword(stemp[37..40])

--	printf(1,"Name RVA: %04xh\n",EDT[5])
--	printf(1,"Ordinal base: %d\n",EDT[6])
--	printf(1,"Address table entries: %d\n",EDT[7])
--	printf(1,"Name pointers: %d\n",EDT[8])
--	printf(1,"Address table RVA: %04x\n",EDT[9])
--	printf(1,"Name table RVA: %04x\n",EDT[10])
--	printf(1,"Ordinal table RVA: %04x\n",EDT[11])
	
		--if seek(fn,rva_to_filepointer(EDT[9],SECT)) then end if
		EAT = repeat(0,EDT[7])
		for i=1 to EDT[7] do
			EAT[i] = gendll_dword(get_bytes(fn,4))
			--printf(1,"RVA %d: %08xh\n",{i,EAT[i]})
		end for
	
		ENP = repeat(0,EDT[8])
		EOT = ENP
		ENT = repeat({},EDT[8]+1)

		for i=1 to EDT[8] do
			ENP[i] = gendll_dword(get_bytes(fn,4))
			--poke4(imagePtr+EDT[9]+((i-1)*4), ENP[i])
		end for
	
		for i=1 to EDT[8] do
			EOT[i] = gendll_word(get_bytes(fn,2))
			--printf(1,"Ordinal %d: %d\n",{i,EOT[i]})
			--poke2(imagePtr+EDT[11]+((i-1)*2), EOT[i])
		end for
	
		--if seek(fn,rva_to_filepointer(EDT[10],SECT)) then end if
		for i=1 to EDT[8]+1 do
			while 1 do
				temp = getc(fn)
				if temp=0 then
					exit
				end if
				ENT[i] &= temp
			end while
			-- Remove name decorations ("_foo@8")
			if length(ENT[i])>=4 then
				if (ENT[i][1]='_') then
					if (ENT[i][length(ENT[i])-1]='@') then
						ENT[i] = ENT[i][2..length(ENT[i])-2]
					elsif (ENT[i][length(ENT[i])-2]='@') then
						ENT[i] = ENT[i][2..length(ENT[i])-3]
					end if
				end if
			end if
			--printf(1,"Name: (%04xh) "&ENT[i]&"\n",where(fn))
			if i=1 then
				--poke(imagePtr+EDT[5], ENT[i]&0)
			else
				--poke(imagePtr+EDT[10], ENT[i]&0)
			end if
		end for
	
	end if



--#################### Process import table ####################	

	if impTblSize>0 then

		if seek(fn, rva_to_filepointer(impTblAddr, SECT)) then end if
		
		stemp = get_bytes(fn,20)
		IDT = repeat(0,6)
		IDT[IDT_LUT_RVA] = gendll_dword(stemp[1..4])
		IDT[IDT_TIME] = gendll_dword(stemp[5..8])
		IDT[IDT_FWD] = gendll_dword(stemp[9..12])
		IDT[IDT_NAME_RVA] = gendll_dword(stemp[13..16])
		IDT[IDT_ADT_RVA] = gendll_dword(stemp[17..20])

		--printf(1,"Name RVA: %04xh\n",IDT[IDT_NAME_RVA])
		--printf(1,"LUT RVA: %04xh\n",IDT[IDT_LUT_RVA])
		--printf(1,"ADT RVA: %04xh\n",IDT[IDT_ADT_RVA])
		
		if seek(fn, rva_to_filepointer(IDT[IDT_NAME_RVA], SECT)) then end if
		
		temp = 1
		stemp = {}
		-- Get name of DLL to import from
		while temp do
			temp = getc(fn)
			if temp!=0 then
				stemp &= temp
			end if
		end while
		
		temp = 1
		-- If the library isn't loaded then we need to load it now
		if not gendll_lib_is_loaded(stemp) then
			temp = 0
			--puts(1,"Loading "&stemp&"\n")
			--while get_key()=-1 do
			--end while
			implib = open_dll(stemp)
			if implib>0 then
				temp = 1
			end if
		end if
		
		-- Fill in the address table
		if temp=1 then
			if seek(fn, rva_to_filepointer(IDT[IDT_LUT_RVA], SECT)) then end if
			temp = 0
			IDT[IDT_LUT] = {}
			while 1 do
				atemp = gendll_dword(get_bytes(fn,4))
				if atemp = 0 then
					exit
				end if
				IDT[IDT_LUT] &= atemp
			end while
			
			--printf(1,"Found %d entries in lookup table\n",length(IDT[IDT_LUT]))
			
			for i=1 to length(IDT[IDT_LUT]) do
				atemp = IDT[IDT_LUT][i]
				
				if atemp>=#80000000 then
					-- entry is an ordinal
				else
					-- entry is a hint/name
					if seek(fn, rva_to_filepointer(atemp, SECT)) then end if
					temp = gendll_word(get_bytes(fn,2))
					temp = 1
					stemp = {}
					-- Get the name of the function 
					while temp do
						temp = getc(fn)
						if temp!=0 then
							stemp &= temp
						end if
					end while
					--puts(1,"Function name: "&stemp&"\n")
					temp = gendll_func_is_def_byname(stemp)
					if not temp then
						poke4(imagePtr+IDT[IDT_ADT_RVA]+((i-1)*4),			
						      gendll_get_proc_address(implib, stemp))
					else
						poke4(imagePtr+IDT[IDT_ADT_RVA]+((i-1)*4),
						      gendll_get_proc_address(temp, stemp))
					end if
				end if
			end for
							
		end if
	
	end if


	
--#################### Process relocation table ####################	

	if relTblSize>0 then
		if seek(fn, rva_to_filepointer(relTblAddr, SECT)) then end if
	
		relDelta = imagePtr - imageBase
		
		while relTblSize>0 do
			pageRVA = gendll_dword(get_bytes(fn,4))
			relBlkSize = gendll_dword(get_bytes(fn,4))-8
-- BUGFIX 020505: get_bytes(fn,2) should only be called once per iteration		
			for i=1 to floor(relBlkSize/2) do
				temp = gendll_word(get_bytes(fn,2))
				relOffs = and_bits(temp,#0FFF)
				relType = floor(temp/#1000)
				if relType=0 then
					--puts(1,"ABSOLUTE\n")
				elsif relType=1 then
					--puts(1,"HIGH\n")
				elsif relType=2 then
					--puts(1,"LOW\n")
				elsif relType=3 then
					--puts(1,"HIGHLOW\n")
					poke4(imagePtr+pageRVA+relOffs,peek4u(imagePtr+pageRVA+relOffs)+relDelta)
				else
					printf(1,"Reltype = %d\n",relType)
				end if
			end for
			relTblSize -= relBlkSize+8
		end while
	end if


	close(fn)

	
	-- This is where you'd call LibMain... I don't.
	--poke4(LibMain_data,imagePtr+entryRVA)
	--poke4(LibMain_data+4,DLL_PROCESS_ATTACH)
	--call(call_LibMain)
	

	--if peek4u(LibMain_data+4)=0 then
	--	-- Something went wrong
	--	free(imagePtr)
	--	return -1
	--end if

	loaded_dlls = append(loaded_dlls,DLL_TEMPLATE)
	loaded_dlls[length(loaded_dlls)][DLL_BASE_ADDRESS] = imagePtr
	loaded_dlls[length(loaded_dlls)][DLL_EAT] = EAT
	loaded_dlls[length(loaded_dlls)][DLL_ENT] = ENT
	loaded_dlls[length(loaded_dlls)][DLL_EOT] = EOT
	loaded_dlls[length(loaded_dlls)][DLL_DEFS] = {}
	
	
	return length(loaded_dlls)
end function




global procedure close_dll(integer which_dll)
	if which_dll<1 or which_dll>length(loaded_dlls) then
		return
	end if

	--poke4(LibMain_data,loaded_dlls[which_dll][DLL_BASE_ADDRESS])
	--poke4(LibMain_data+4,DLL_PROCESS_DETACH)
	--call(call_LibMain)

	for i=1 to length(loaded_dlls[which_dll][DLL_DEFS]) do
		free(gendll_fundefs[loaded_dlls[which_dll][DLL_DEFS][i]][1])
		gendll_ordinals[loaded_dlls[which_dll][DLL_DEFS][i]] = -1
	end for
	
	free(loaded_dlls[which_dll][DLL_BASE_ADDRESS])
	loaded_dlls[which_dll] = {}
end procedure



global procedure gendll_set_call_mode(integer cm)
	gendll_call_mode = cm
end procedure



global function define_c_func(atom lib, sequence func_name, sequence param_types, atom return_type)

	integer j,next_free_reg
	sequence s,push_code,data_sect,stackfix_code,plist
	atom call_code
	integer parmTotSize
	
	next_free_reg = 1
	
	if lib<=0 then
		return -1
	end if

	s = loaded_dlls[lib]

	j = 0
	for i=2 to length(s[DLL_ENT]) do
		if equal(func_name,s[DLL_ENT][i]) then
			j = i-1
			exit
		end if
	end for
	if j=0 then
		 return -1
	end if
	
	parmTotSize = 0
	push_code = {}
	data_sect = "nop nop nop nop "
	if length(param_types) then
		plist = repeat(0,length(param_types))
	else
		plist = {}
	end if
	
	if length(param_types) then
		for i=length(param_types) to 1 by -1 do
			if param_types[i] = C_CHAR or param_types[i]=C_UCHAR then
				if gendll_call_mode=GENDLL_FASTCALL then
					push_code &= sprintf("mov "&gendll_regs[i]&",[param%02d] ", i)
				else
					push_code &= sprintf("mov al,[param%02d] dec esp mov [esp],al ",i)
					parmTotSize += 1
					plist[i] = 1
				end if
			elsif param_types[i] = C_SHORT or param_types[i]=C_USHORT then
				if gendll_call_mode=GENDLL_FASTCALL then
					push_code &= sprintf("mov "&gendll_regs[i]&",[param%02d] ", i)
				else
					push_code &= sprintf("push word ptr [param%02d] ",i)
					parmTotSize += 2
					plist[i] = 2
				end if
			elsif param_types[i] = C_INT or param_types[i]=C_UINT or param_types[i]=C_FLOAT then
				if gendll_call_mode=GENDLL_FASTCALL then
					push_code &= sprintf("mov "&gendll_regs[i]&",[param%02d] ", i)
				else
					push_code &= sprintf("push dword ptr [param%02d] ",i)
					parmTotSize += 4
					plist[i] = 4
				end if
			end if
-- BUGFIX 020512: Shouls always be 4 nop:s			
			data_sect &= "nop nop nop nop "
		end for
	end if
	
	--puts(1,data_sect)
	--puts(1,push_code)
	
	stackfix_code = {}
	if gendll_call_mode = GENDLL_CDECL then
		stackfix_code = sprintf("add esp,%d ",parmTotSize)
	end if
	
	call_code = get_asm(
	data_sect&
	"pusha "&		-- Save all registers, just to make sure. Some functions destroy registers without restoring them
	push_code&
	"mov ebp,funaddr "&
	"call near ebp "&
	"mov [result],eax "&
	stackfix_code&
	"popa "&
	"ret")
	resolve_param("funaddr",s[DLL_BASE_ADDRESS]+s[DLL_EAT][j])
	resolve_param("result",call_code)
	if length(param_types) then
		for i=1 to length(param_types) do
			resolve_param(sprintf("param%02d",i),call_code+(i*4))
		end for
	end if
	
	gendll_fundefs = append(gendll_fundefs, {call_code, call_code+(length(param_types)*4)+4,plist,return_type})
	gendll_ordinals &= s[DLL_EOT][j]
	loaded_dlls[lib][DLL_DEFS] &= length(gendll_fundefs)
	
	return length(gendll_fundefs)
end function




global function define_c_proc(atom lib, sequence proc_name, sequence param_types)
	return define_c_func(lib, proc_name, param_types, C_INT)
end function



global function c_func(integer fun_id, sequence params)

	sequence s

	s = gendll_fundefs[fun_id]

	poke4(s[1]+4,params)
	call(s[2])

	if s[4]=C_CHAR then
		return peek(s[1])
	elsif s[4]=C_INT then
		return peek4s(s[1])
	elsif s[4]=C_UINT then
		return peek4u(s[1])
	end if

	return 0
end function



global procedure c_proc(integer proc_id, sequence params)

	sequence s

	s = gendll_fundefs[proc_id]

	poke4(s[1]+4,params)
	call(s[2])

end procedure

