;######################## KERNEL32.DLL replacement for DOS (+Causeway) ######################

;
; Compile with MASM:
;
;   ml /c /coff dosknl32.asm
;
;   link /SUBSYSTEM:WINDOWS /DLL /DEF:dosknl32.def /OUT:dosknl32.dll dosknl32.obj
;
;
; /Mic, 2002
; stabmaster_@hotmail.com
;


; Functions that are working (more or less):
;
;	GetLastError
;	CloseHandle		(only supports files - no consoles, pipes, mailslots and whatnot)
;	CreateFile		(only supports files)
;	LocalAlloc
;	LocalFree
;	ReadFile		(only supports files)
;	SetConsoleCursorPosition
;	SetLastError
;	VirtualAlloc
;	VirtualFree
;	WriteConsole		(always writes to the screen)
;	WriteFile		(only supports files)
;	_lclose
;	_lcreat
;	_llseek	
;	_lopen
;	_lread
;	_lwrite
;
;
; Functions that are available, but only as stubs:
;
;	EnterCriticalSection
;	ExitProcess
;	GetACP
;	GetOEMCP
;	GetCurrentProcess
;	GetCurrentThreadId
;	GetFileType
;	GetVersion
;	HeapAlloc
;	HeapCreate
;	HeapDestroy
;	HeapFree
;	HeapReAlloc
;	InitializeCriticalSection
;	LCMapString
;	LCMapStringW
;	LeaveCriticalSection
;	RtlUnwind
;	TlsAlloc
;	TlsGetValue
;	TlsFree
;	TlsSetValue


    .386
    .model flat, stdcall
    option casemap :none   ; case sensitive


;#########################################################################

    include windows.inc

;#########################################################################

    szText MACRO Name, Text:VARARG
      LOCAL lbl
        jmp lbl
          Name db Text,0
        lbl:
      ENDM

    m2m MACRO M1, M2
      push M2
      pop  M1
    ENDM

    return MACRO arg
      mov eax, arg
      ret
    ENDM

;##########################################################################

.data?

iTemp		dd ?


.data

FILEMODE_R	equ 0
FILEMODE_W	equ 1
FILEMODE_RW	equ 2

HFILE_ERROR	equ -1


dummyString	db	'swizzlesticks',0,0,0,0
lineFeed	db	13,10,'$'
dwLastError	dd	0


;##########################################################################

.code

LibMain PROC hInstDLL:DWORD, reason:DWORD, unused:DWORD


        .if reason == 1		;DLL_PROCESS_ATTACH

            return TRUE
            ; -----------------------------
            ; If error at startup, return 0
            ; System will abort loading DLL
            ; -----------------------------

        .elseif reason == 0	;DLL_PROCESS_DETACH

        .elseif reason == 2	;DLL_THREAD_ATTACH

        .elseif reason == 3	;DLL_THREAD_DETACH
            
        .endif

        return FALSE

LibMain ENDP



;######################## Process/Thread management ########################


TerminateProcess PROC hProcess:DWORD,uExitCode:DWORD
	xor eax,eax
	ret
TerminateProcess ENDP


ExitProcess PROC uExitCode:DWORD
	mov eax,uExitCode
	ret
ExitProcess ENDP


GetCurrentProcess PROC
	xor eax,eax
	ret
GetCurrentProcess ENDP


GetCurrentThreadId PROC
	xor eax,eax
	ret
GetCurrentThreadId ENDP



;######################## DLL functions ########################

GetModuleFileNameA PROC hModule:DWORD,lpFileName:DWORD,nSize:DWORD
	push ecx
	mov ecx,14
	cmp nSize,14
	jge sizeOK
		mov ecx,nSize
	sizeOK:
	push esi
	push edi
	mov esi,OFFSET dummyString
	mov edi,lpFileName
	rep movsb
	pop edi
	pop esi
	pop ecx
	ret
GetModuleFileNameA ENDP


GetProcAddress PROC hModule:DWORD,lpProcName:DWORD
	xor eax,eax
	ret
GetProcAddress ENDP


LoadLibraryA PROC lpLibFileName:DWORD
	xor eax,eax
	ret
LoadLibraryA ENDP


LoadLibraryEx PROC lpLibFileName:DWORD,hFile:DWORD,dwFlags:DWORD
	xor eax,eax
	ret
LoadLibraryEx ENDP




;######################## Thread Local Storage management ########################

TlsAlloc PROC
	mov eax,0FFFFFFFFh
	ret
TlsAlloc ENDP


TlsFree PROC dwTlsIndex:DWORD
	xor eax,eax
	ret
TlsFree ENDP


TlsGetValue PROC dwTlsIndex:DWORD
	xor eax,eax
	ret
TlsGetValue ENDP


TlsSetValue PROC dwTlsIndex:DWORD,lpTlsValue:DWORD
	xor eax,eax
	ret
TlsSetValue ENDP




;######################## File handling functions ########################

_lopen PROC lpPathName:DWORD,iReadWrite:DWORD
	LOCAL hFile:DWORD
	
	pusha
	
	mov edx,lpPathName
	mov ebx,iReadWrite
	
	.if ebx==OF_READ
		mov al,FILEMODE_R
	.elseif ebx==OF_WRITE
		mov al,FILEMODE_W
	.elseif ebx==OF_READWRITE
		mov al,FILEMODE_RW
	.else
		popa
		return HFILE_ERROR
	.endif

	mov ah,03Dh
	int 21h
	
	jnc LO_open_ok
		mov dwLastError,ERROR_FILE_NOT_FOUND
		popa
		return HFILE_ERROR
	
	LO_open_ok:
	and eax,0FFFFh
	mov hFile,eax
	popa
	return hFile
_lopen ENDP


_lcreat PROC lpPathName:DWORD,iAttribute:DWORD
	LOCAL hFile:DWORD
	
	pusha
	
	mov ecx,iAttribute
	mov edx,lpPathName
	mov ah,03Ch
	int 21h
	jnc LC_create_ok
		popa
		return HFILE_ERROR
	LC_create_ok:
	and eax,0FFFFh
	mov hFile,eax
	popa
	return hFile
_lcreat ENDP


_llseek PROC hFile:DWORD,lOffset:DWORD,iOrigin:DWORD
	LOCAL dwNewPos:DWORD
	
	pusha
	
	mov ecx,lOffset
	mov edx,ecx
	shr ecx,16
	and edx,0FFFFh
	
	mov ebx,iOrigin
	.if ebx==FILE_BEGIN
		mov al,0
	.elseif ebx==FILE_CURRENT
		mov al,1
	.elseif ebx==FILE_END
		mov al,2
	.else
		popa
		return HFILE_ERROR
	.endif
	
	mov ah,042h
	int 21h
	jnc LLS_seek_ok
		popa
		return HFILE_ERROR
	LLS_seek_ok:
	and eax,0FFFFh
	shl edx,16
	or eax,edx
	mov dwNewPos,eax
	popa
	return dwNewPos
_llseek ENDP


_lread PROC hFile:DWORD,lpBuffer:DWORD,uBytes:DWORD
	LOCAL uBytesRead:DWORD
	
	pusha
	
	mov ebx,hFile
	mov ecx,uBytes
	mov edx,lpBuffer
	mov ah,03Fh
	int 21h
	jnc LR_read_ok
		popa
		return HFILE_ERROR
	LR_read_ok:
	mov uBytesRead,eax
	popa
	return uBytesRead
_lread ENDP


_lwrite PROC hFile:DWORD,lpBuffer:DWORD,uBytes:DWORD
	LOCAL uBytesWritten:DWORD
	
	pusha
	
	mov ebx,hFile
	mov ecx,uBytes
	mov edx,lpBuffer
	mov ah,040h
	int 21h
	jnc LW_write_ok
		popa
		return HFILE_ERROR
	LW_write_ok:
	mov uBytesWritten,eax
	popa
	return uBytesWritten
_lwrite ENDP
	
	
_lclose PROC hFile:DWORD
	pusha

	mov ebx,hFile
	mov ah,03Eh
	int 21h
	jnc LC_close_ok
		popa
		return HFILE_ERROR
	LC_close_ok:
	popa
	return 0
_lclose ENDP


CloseHandle PROC hObject:DWORD
	invoke _lclose,hObject
	.if eax==HFILE_ERROR
		return 0
	.endif
	return 1
CloseHandle ENDP


; HANDLE CreateFile
; Working
;
CreateFileA PROC lpFileName:DWORD,dwDesiredAccess:DWORD,dwShareMode:DWORD,lpSecurityAttributes:DWORD,
                 dwCreationDistrobution:DWORD,dwFlagsAndAttributes:DWORD,hTemplateFile:DWORD

	LOCAL hFile:DWORD,dwAccess:DWORD
	
	pusha
	
	and dwFlagsAndAttributes,0FFFFh
	
	mov ecx,GENERIC_READ
	or ecx,GENERIC_WRITE
	
	mov ebx,dwDesiredAccess
	.if ebx==GENERIC_READ
		mov dwAccess,FILEMODE_R
	.elseif ebx==GENERIC_WRITE
		mov dwAccess,FILEMODE_W
	.elseif ebx==ecx
		mov dwAccess,FILEMODE_RW
	.else
		popa
		return INVALID_HANDLE_VALUE
	.endif

	mov ebx,dwCreationDistrobution
	.if ebx==CREATE_NEW
		invoke _lopen,lpFileName,dwAccess
		.if eax==HFILE_ERROR
			.if dwFlagsAndAttributes==FILE_ATTRIBUTE_NORMAL
				invoke _lcreat,lpFileName,0
			.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_HIDDEN
				invoke _lcreat,lpFileName,2
			.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_READONLY
				invoke _lcreat,lpFileName,1
			.else
				popa
				return INVALID_HANDLE_VALUE
			.endif
			.if eax==HFILE_ERROR
				popa
				return INVALID_HANDLE_VALUE
			.else
				mov hFile,eax
				popa
				return hFile
			.endif
		.else
			mov dwLastError,ERROR_ALREADY_EXISTS
			invoke _lclose,eax
			popa
			return INVALID_HANDLE_VALUE
		.endif

	.elseif ebx==CREATE_ALWAYS
		.if dwFlagsAndAttributes==FILE_ATTRIBUTE_NORMAL
			invoke _lcreat,lpFileName,0
		.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_HIDDEN
			invoke _lcreat,lpFileName,2
		.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_READONLY
			invoke _lcreat,lpFileName,1
		.else
			popa
			return INVALID_HANDLE_VALUE
		.endif

		.if eax==HFILE_ERROR
			popa
			return INVALID_HANDLE_VALUE
		.else
			mov hFile,eax
			popa
			return hFile
		.endif

	.elseif ebx==OPEN_EXISTING
		invoke _lopen,lpFileName,dwAccess
		.if eax==HFILE_ERROR
			popa
			return INVALID_HANDLE_VALUE
		.else
			mov hFile,eax
			popa
			return hFile
		.endif

	.elseif ebx==OPEN_ALWAYS
		invoke _lopen,lpFileName,dwAccess
		.if eax==HFILE_ERROR
			.if dwFlagsAndAttributes==FILE_ATTRIBUTE_NORMAL
				xor eax,eax
				invoke _lcreat,lpFileName,eax
			.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_HIDDEN
				invoke _lcreat,lpFileName,2
			.elseif dwFlagsAndAttributes==FILE_ATTRIBUTE_READONLY
				invoke _lcreat,lpFileName,1
			.else
				xor eax,eax
				invoke _lcreat,lpFileName,eax
				;popa
				;return INVALID_HANDLE_VALUE
			.endif
			.if eax==HFILE_ERROR
				popa
				return INVALID_HANDLE_VALUE
			.else
				mov hFile,eax
				popa
				return hFile
			.endif
		.else
			mov hFile,eax
			popa
			return hFile
		.endif
	
	.else
		mov dwLastError,ERROR_NOT_SUPPORTED
		popa
		return INVALID_HANDLE_VALUE
	.endif
	
	popa
	return INVALID_HANDLE_VALUE
CreateFileA ENDP



ReadFile PROC hFile:DWORD,lpBuffer:DWORD,nNumberOfBytesToRead:DWORD,lpNumberOfBytesRead:DWORD,lpOverlapped:DWORD
	pusha
	invoke _lread,hFile,lpBuffer,nNumberOfBytesToRead
	.if eax==HFILE_ERROR
		popa
		return 0
	.endif
	mov ebx,lpNumberOfBytesRead
	mov [ebx],eax
	popa
	return 1
ReadFile ENDP


WriteFile PROC hFile:DWORD,lpBuffer:DWORD,nNumberOfBytesToWrite:DWORD,lpNumberOfBytesWritten:DWORD,lpOverlapped:DWORD
	pusha
	invoke _lwrite,hFile,lpBuffer,nNumberOfBytesToWrite
	.if eax==HFILE_ERROR
		popa
		return 0
	.endif
	mov ebx,lpNumberOfBytesWritten
	mov [ebx],eax
	popa
	return 1
WriteFile ENDP


; DWORD GetFileType
; Stub
;
GetFileType PROC hFile:DWORD
	mov eax,FILE_TYPE_UNKNOWN
	ret
GetFileType ENDP



;######################## Console functions ########################

; BOOL WriteConsole
; Working
; Equivalent to puts(stdout,"...")
;
WriteConsoleA PROC hConsoleOutput:DWORD,lpBuffer:DWORD,nNumberOfCharsToWrite:DWORD,
                  lpNumberOfCharsWritten:DWORD,lpReserved:DWORD

	pusha
	
	.if lpReserved!=NULL
		popa
		return 0
	.endif
	
	mov edx,lpBuffer
	mov eax,nNumberOfCharsToWrite
	mov ebx,lpNumberOfCharsWritten
	mov [ebx],eax
	mov bl,[edx+eax]
	push ebx
	mov bl,'$'
	mov [edx+eax],bl
	mov ah,09h
	int 21h
	pop ebx
	mov edx,lpBuffer
	mov eax,nNumberOfCharsToWrite
	mov [edx+eax],bl
	
	popa
	return 1
WriteConsoleA ENDP



SetConsoleCursorPosition PROC hConsoleOutput:DWORD,dwCursorPosition:DWORD
	pusha
	mov ecx,dwCursorPosition
	mov edx,ecx
	and edx,0FFh		; Put X coordinate in DL
	shr ecx,8
	and ecx,0FF00h		
	or edx,ecx		; Put Y coordinate in DH
	mov bh,0
	mov ah,02h
	int 10h
	popa
	return 1
SetConsoleCursorPosition ENDP


;######################## Synchronization functions ########################

InitializeCriticalSection PROC lpCriticalSection:DWORD
	xor eax,eax
	ret
InitializeCriticalSection ENDP


EnterCriticalSection PROC lpCriticalSection:DWORD
	xor eax,eax
	ret
EnterCriticalSection ENDP


LeaveCriticalSection PROC lpCriticalSection:DWORD
	xor eax,eax
	ret
LeaveCriticalSection ENDP


DeleteCriticalSection PROC lpCriticalSection:DWORD
	xor eax,eax
	ret
DeleteCriticalSection ENDP



;######################## Codepage functions ########################

GetCPInfo PROC CodePage:DWORD,lpCPInfo:DWORD
	xor eax,eax
	ret
GetCPInfo ENDP


GetACP PROC 
	mov eax,1252			; 1252 = Win 3.1 US/Western European ANSI Code Page
	ret
GetACP ENDP


GetOEMCP PROC
	mov eax,437			; 437 = US MS-DOS
	ret
GetOEMCP ENDP



;######################## Memory management ########################

VirtualAlloc PROC lpAddress:DWORD,dwSize:DWORD,flAllocationType:DWORD,flProtect:DWORD
	LOCAL dwLinearAddress:DWORD
	
	pusha
	
	mov ecx,dwSize
	mov ax,0FF11h
	int 31h			; CauseWay function FF11h: GetMemLinear32
	jc VA_alloc_failed
	mov dwLinearAddress,esi	
	jmp VA_return
	
	VA_alloc_failed:
	mov dwLinearAddress,0
	
	VA_return:
	popa
	mov eax,dwLinearAddress
	ret
VirtualAlloc ENDP


VirtualFree PROC lpAddress:DWORD,dwSize:DWORD,dwFreeType:DWORD
	pusha
	
	mov esi,lpAddress
	mov ax,0FF15h
 	int 31h			; FF15h: ReleaseMemLinear32
	jc VF_failed
	popa
	mov eax,1
	ret
	
	VF_failed:
	popa
	xor eax,eax
	ret
VirtualFree ENDP


LocalAlloc PROC uFlags:DWORD,uSize:DWORD
	LOCAL dwLinearAddress:DWORD
	
	pusha
	
	mov ecx,uSize
	mov ax,0FF11h
	int 31h			; CauseWay function FF11h: GetMemLinear32
	jc LA_alloc_failed
	mov dwLinearAddress,esi	
	jmp LA_return
	
	LA_alloc_failed:
	mov dwLinearAddress,0
	
	LA_return:
	popa
	mov eax,dwLinearAddress
	ret
LocalAlloc ENDP


LocalFree PROC hMem:DWORD
	pusha
	
	cmp hMem,0
	jne LF_ok
	popa
	xor eax,eax
	ret
	
	LF_ok:
	mov esi,hMem
	mov ax,0FF15h
 	int 31h			; FF15h: ReleaseMemLinear32
	jc LF_failed
	popa
	mov eax,0
	ret
	
	LF_failed:
	popa
	mov eax,1
	ret
LocalFree ENDP


HeapAlloc PROC hHeap:DWORD,dwFlags:DWORD,dwBytes:DWORD
	xor eax,eax
	ret
HeapAlloc ENDP


HeapReAlloc PROC hHeap:DWORD,dwFlags:DWORD,lpMem:DWORD,dwBytes:DWORD
	xor eax,eax
	ret
HeapReAlloc ENDP


HeapCreate PROC flOptions:DWORD,dwInitialSize:DWORD,dwMaximumSize:DWORD
	xor eax,eax
	ret
HeapCreate ENDP


HeapDestroy PROC hHeap:DWORD
	xor eax,eax
	ret
HeapDestroy ENDP


HeapFree PROC hHeap:DWORD,dwFlags:DWORD,lpMem:DWORD
	xor eax,eax
	ret
HeapFree ENDP



;######################## String functions ########################

LCMapStringA PROC Locale:DWORD,dwMapFlags:DWORD,lpSrcStr:DWORD,cchSrc:DWORD,lpDestStr:DWORD,cchDest:DWORD
	xor eax,eax
	ret
LCMapStringA ENDP


LCMapStringW PROC Locale:DWORD,dwMapFlags:DWORD,lpSrcStr:DWORD,cchSrc:DWORD,lpDestStr:DWORD,cchDest:DWORD
	xor eax,eax
	ret
LCMapStringW ENDP


GetStringTypeA PROC Locale:DWORD,dwInfoType:DWORD,lpSrcStr:DWORD,cchSrc:DWORD,lpCharType:DWORD
	xor eax,eax
	ret
GetStringTypeA ENDP


GetStringTypeW PROC Locale:DWORD,dwInfoType:DWORD,lpSrcStr:DWORD,cchSrc:DWORD,lpCharType:DWORD
	xor eax,eax
	ret
GetStringTypeW ENDP


GetEnvironmentStringsW PROC
	mov eax,OFFSET dummyString
	ret
GetEnvironmentStringsW ENDP


GetEnvironmentStrings PROC
	mov eax,OFFSET dummyString
	ret
GetEnvironmentStrings ENDP


FreeEnvironmentStringsA PROC lpzEnvironmentBlock:DWORD
	xor eax,eax
	ret
FreeEnvironmentStringsA ENDP
	

FreeEnvironmentStringsW PROC lpzEnvironmentBlock:DWORD
	xor eax,eax
	ret
FreeEnvironmentStringsW ENDP


MultiByteToWideChar PROC CodePage:DWORD,dwFlags:DWORD,lpWideCharString:DWORD,cchWideChar:DWORD,
		         lpMultiByteString:DWORD,cchMultiByte:DWORD,lpDefaultChar:DWORD,lpUsedDefaultChar:DWORD
	xor eax,eax
	ret
MultiByteToWideChar ENDP


WideCharToMultiByte PROC CodePage:DWORD,dwFlags:DWORD,lpWideCharString:DWORD,cchWideChar:DWORD,
		         lpMultiByteString:DWORD,cchMultiByte:DWORD,lpDefaultChar:DWORD,lpUsedDefaultChar:DWORD
	xor eax,eax
	ret
WideCharToMultiByte ENDP



;######################## Misc. functions ########################

GetCommandLineA PROC
	mov eax,OFFSET dummyString
	ret
GetCommandLineA ENDP


RtlUnwind PROC
	xor eax,eax
	ret
RtlUnwind ENDP


SetHandleCount PROC uNumber:DWORD
	mov eax,uNumber
	ret
SetHandleCount ENDP


GetStartupInfoA PROC lpStartupInfo:DWORD
	mov eax,lpStartupInfo
	push ebx
	xor ebx,ebx
	mov [eax],ebx
	mov [eax+4],ebx
	mov [eax+8],ebx
	mov [eax+12],ebx
	pop ebx
	ret
GetStartupInfoA ENDP


GetStdHandle PROC nStdHandle:DWORD
	mov eax,INVALID_HANDLE_VALUE
	ret
GetStdHandle ENDP


GetVersion PROC
	mov eax,0104h			; Make the caller belive this is Win95 (4.1)
	ret
GetVersion ENDP


GetLastError PROC
	return dwLastError
GetLastError ENDP


SetLastError PROC dwErrCode:DWORD
	mov eax,dwErrCode
	mov dwLastError,eax
	ret
SetLastError ENDP





; ##########################################################################

End LibMain
